/**
 * Bruce Land and Hunter Adams (vha3@cornell.edu)
 * ======================================================
 * simmple serial command interface
 * ======================================================
 * 
 * Protothreads v1.3
 * Serial console on GPIO 0 and 1 for debugging
 * and for synth conrol
 * 
 * =====================================================
 * 
 * There are two threads running 
 * 
 * ==========
 * Core0:
 * 
 * ==========
 * Core1:
 * 
 * ---- Handles the serial interface for setting parameters
 * 
 * -- A play thread which
 * ---- sequences some blinking pixels
 * 
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "pico/stdlib.h"
#include "hardware/pio.h"
#include "hardware/adc.h"
#include "hardware/irq.h"
#include "hardware/spi.h"

// ==========================================
// === hardware and protothreads globals
// ==========================================
#include "hardware/sync.h"
#include "hardware/timer.h"
#include "pico/multicore.h"
#include "string.h"
// protothreads header
#include "pt_cornell_rp2040_v1_3.h"

int titlecolor = 231, titlebkgd = 238 ;
char titlestr[64] = "Serial command example, ece4760 Cornell University, 2024";
char title_style ;
char data_string[64], data_string2[64];
char style_str[20] = "bold";
bool blinky_flag;
float num1, num2, result ;
bool calc_flag ;
char op ;
// =========================================
// === serial formatting macros
// =========================================
#define new_screen printf("\14")
// top of screen is line 1
#define cursor_pos(line,col) printf("\033[%d;%dH", line, col)
// font control
#define Boldface printf("\33[1m")
#define Italic  printf("\33[23m")
#define Normaltext printf("\33[0m") // (undoes all text highlighting and color)
#define doublewidth printf("\33#6")
// hardcode color connrol
#define Greentext printf("\33[32m")
#define BoldGreentext printf("\33[1;32m")
//
#define SaveCursor printf("\33[s");
#define RestoreCursor printf("\33[u");
#define InvisibleCur printf("\33[?25l");
#define VisibleCur printf("\33[?25h");

// === 8 color mode ===
#define BLACK 30
#define RED 31
#define GREEN 32
#define YELLOW 33
#define BLUE 34
#define MAGENTA 35
#define CYAN 36
#define WHITE 37
#define SetTextColor8(c) printf("\33[%2dm", c)
#define SetBkgdColor8(c) printf("\33[%2dm", c+10)
// === 256 MODE COLOR  === see 
// https://gist.github.com/fnky/458719343aabd01cfb17a3a4f7296797
#define SetTextColor(c) printf("\33[38;5;%dm", c);
#define SetBkgdColor(c) printf("\33[48;5;%dm", c);



// ==================================================
// === print the parameters and results
// ==================================================
// 

// shoot all the paramaters to the serial
void print_all(void){
    new_screen ;
    cursor_pos(1,5);
    SetTextColor(titlecolor); SetBkgdColor(titlebkgd);
    if(title_style=='n') Normaltext;
    else if(title_style=='b') Boldface;
    else if(title_style=='d') doublewidth;
    printf("%s", titlestr);
    Normaltext;


    cursor_pos(3,0); 
    printf( "titlecolors = "); 
      SetTextColor(195); 
      printf("%d,  ", titlecolor);
      //SetTextColor(titlebkgd);
      //SetBkgdColor((titlebkgd+200)%256); 
      printf("%d  ", titlebkgd);
      Normaltext ;

    cursor_pos(4,0);  
     printf( "titlestyle =");
    SetTextColor(195); 
    printf( " %s  ", style_str);
     Normaltext ;
   

    cursor_pos(6,0); 
    if(calc_flag) {
      printf( "calc ( %s ) = %8.5f   ", data_string2, result);
      calc_flag = false ;
    }

    cursor_pos(10,0); 
    //commands
    SetTextColor8(GREEN);
    printf("\n\r***commands***\n\r");          
    printf(" title color forgnd bkgd -- forgnd bkgd: 0 to 255\n\r");
    printf("       style type -- type={normal, bold, double}\n\r");
    printf("       text string  -- up to 50 chars\n\r");
    //
    printf("\n\r calc number operator number -- operator={+, -, *, /}\n\r");
    //
    printf("\n\r blinky on/off\n\r");
    //
    
    //
    printf("\n\r");
    Normaltext ;

}


///==============================================
// === interactive input
// ==============================================

static PT_THREAD (protothread_serial(struct pt *pt))
{
  PT_BEGIN(pt);
    
    static char serial_buffer[40];
    static char cmd[16], arg1[16], arg2[16], arg3[16], whole_cmd[64] ;
    static char* token ;
    static int bad_cmd ;
    float farg ;
    int arg ;

    //
    print_all() ;
    
    //
    while(1) {

      // print prompt
        cursor_pos(23,0); 
        printf("                    ");
        cursor_pos(23,0); Boldface; 
        printf("cmd>                         ");
        cursor_pos(23,5);
        Normaltext;
        
        // spawn a thread to do the non-blocking serial read
         serial_read ;

        // store a copy for error reporting
        strcpy(whole_cmd, pt_serial_in_buffer) ;
        // set up tokeizer
         char* rest = pt_serial_in_buffer ;

        // tokenize
        token = strtok_r(rest, " ", &rest);
        strcpy(cmd, token) ;
        // store a copy for error reporting
         strcpy(data_string2, rest) ;

        //token = strtok(NULL, "  ");
        token = strtok_r(rest, " ", &rest);
        strcpy(arg1, token) ;
        // store a copy for error reporting
        strcpy(data_string, rest) ;

        //token = strtok(NULL, "  ");
        token = strtok_r(rest, " ", &rest);
        strcpy(arg2, token) ;
        
       // token = strtok(NULL, "  ");
        token = strtok_r(rest, " ", &rest);
        strcpy(arg3, token) ;
      
      
      // parse by command
        // title command
        if(strcmp(cmd,"title")==0){
          // color option
          if(strcmp(arg1,"color")==0){
            sscanf(arg2,"%d", &titlecolor);
            sscanf(arg3,"%d", &titlebkgd);
          }
          // style option
          if(strcmp(arg1,"style")==0){
            if(strcmp(arg2,"normal")==0){
              title_style = 'n' ;
              strcpy(style_str, "normal");
            }
            else if(strcmp(arg2,"bold")==0){
              title_style = 'b' ;
              strcpy(style_str, "bold");
            }
             else if(strcmp(arg2,"double")==0){
              title_style = 'd' ;
              strcpy(style_str, "double width");
            }   
            else bad_cmd = 2 ;         
          }
          // text change option
          if(strcmp(arg1,"text")==0){
            strcpy(titlestr, data_string) ;
          }
        }

        // turn on stupid blinking characters
        else if(strcmp(cmd,"blinky")==0){
            if(strcmp(arg1,"on")==0) blinky_flag = true ;
            else blinky_flag = false ;
        }

        
        // elementary 4 funcdtion calc
        // parse differently becuase there maybe no spaces bwteeen
        // operator and operands.
        else if(strcmp(cmd,"calc")==0){
            char * ptr;
            // find and set operator
            if(ptr = strchr(data_string2, '+')) op = '+' ;
            else if(ptr = strchr(data_string2, '-')) op = '-' ;
            else if(ptr = strchr(data_string2, '*')) op = '*' ;
            else if(ptr = strchr(data_string2, '/')) op = '/' ;
            else bad_cmd = 3;
            // delete operator from string
            *ptr = ' ';
            // get the two operands from the modifed stirng
             sscanf(data_string2,"%f %f", &num1, &num2);
            //do the op
            if(op == '+') result = num1 + num2 ;
            if(op == '-') result = num1 - num2 ;
            if(op == '*') result = num1 * num2 ;
            if(op == '/') result = num1 / num2 ;
            // put op back into string for printing
            *ptr = op ;
            calc_flag = true;
        }

        //
        else {
            bad_cmd = 1 ;
        }
        //
        if(!bad_cmd) {
          print_all();
        }
        else {
          // error reporting
           cursor_pos(22,0); 
           if(bad_cmd == 1)  printf("***Bad command: %s           ", whole_cmd); 
           if(bad_cmd == 2)  printf("***Bad title style: %s        ", data_string); 
           if(bad_cmd == 3)  printf("***Bad arithmetic specified: %s        ", data_string2); 
        }
        bad_cmd = 0;
      

      // NEVER exit while
    } // END WHILE(1)
  PT_END(pt);
} // timer thread

// ==================================================
// === blink the pixels, if turned on
// ==================================================

#define xpos 70
#define ypos 10
static PT_THREAD (protothread_play(struct pt *pt))
{
    PT_BEGIN(pt);

      while(1) {
        PT_YIELD_UNTIL(pt, blinky_flag);
        //
        PT_YIELD_usec(500000) ;
        SaveCursor;
        InvisibleCur;
        cursor_pos(ypos, xpos); 
       // SetTextColor(199); SetBkgdColor(196);
       SetTextColor8(GREEN); SetBkgdColor8(RED);
        printf("**");
        //cursor_pos(ypos+1, xpos);
        //printf("**");
        RestoreCursor; 
        VisibleCur ;
        //
        PT_YIELD_usec(500000) ;
        SaveCursor;
        InvisibleCur;
         cursor_pos(ypos, xpos); 
         SetTextColor(0); SetBkgdColor(0);
        printf("  ");
        // cursor_pos(ypos+1, xpos); 
        //printf("  ");
        RestoreCursor;  
        VisibleCur;
        
        //
        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // play thread

// ========================================
// === core 1 main -- started in main below
// ========================================
void core1_main(){ 

    //  === add threads  ====================
    // for core 1
    pt_add_thread(protothread_serial) ;
    pt_add_thread(protothread_play) ;
    //
    // === initalize the scheduler ==========
    pt_schedule_start ;
    // NEVER exits
    // ======================================
}

// ========================================
// === core 0 main
// ========================================
int main(){

  // start the serial i/o
  stdio_init_all() ;
  // announce the threader version on system reset
  printf("\n\rProtothreads RP2040 v1.3 two-core, priority\n\r");

  // start core 1 threads
  multicore_reset_core1();
  multicore_launch_core1(&core1_main);

  // === config threads ========================
  // for core 0
  //pt_add_thread(protothread_toggle25);
  //
  // === initalize the scheduler ===============
  pt_schedule_start ;
  // NEVER exits
  // ===========================================
} // end main